/* "eqcs_bitset_rep.cc"
 * Copyright (C) 1999 Peter Belkner <pbelkner@berlin.snafu.de>
 *
 * This file is part of the eqcs library. This library is free
 * software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License as published by the Free Software
 * Foundation; either version 2, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; see the file COPYING.  If not, write to the
 * Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#ifdef __GNUG__
#pragma implementation
#endif

#include <assert.h>
#include <iostream.h>
#include <algorithm>
#include "eqcs_sep.h"
#include "eqcs_bitset_rep.h"

/*****************************************************************************\
*	bitset representation													  *
\*****************************************************************************/

EqcsBitsetRep::EqcsBitsetRep(int nbits) : _rep(dim(nbits))
{
	_nbits = nbits;
	_nbitson = 0;

	for (rep_t::iterator p = _rep.begin(); p != _rep.end(); ++p)
		*p = 0;
}

EqcsBitsetRep::~EqcsBitsetRep()
	{}

int EqcsBitsetRep::nbits() const
	{ return _nbits; }

int EqcsBitsetRep::nbitson() const
	{ return _nbitson; }

void EqcsBitsetRep::set(int bit)
{
	assert(0 <= bit && bit < _nbits);

	slot_t mask = 1ul << (bit % BITS_PER_SLOT);
	int i = bit / BITS_PER_SLOT;

	if (0ul == (_rep[i] & mask)) {
		_rep[i] |= mask;
		++_nbitson;
	}
}

void EqcsBitsetRep::reset(int bit)
{
	assert(0 <= bit && bit < _nbits);

	slot_t mask = 1ul << (bit % BITS_PER_SLOT);
	int i = bit / BITS_PER_SLOT;

	if (0ul != (_rep[i] & mask)) {
		_rep[i] &= ~mask;
		--_nbitson;
	}
}

bool EqcsBitsetRep::test(int bit) const
{
	assert(0 <= bit);

	if (bit < _nbits) {
		slot_t mask = 1ul << (bit % BITS_PER_SLOT);
		int i = bit / BITS_PER_SLOT;

		return 0ul != (_rep[i] & mask);
	}
	else
		return false;
}

int EqcsBitsetRep::cmp(const EqcsBitsetRep *bitset) const
{
	int i = 0;

	if (_rep.size() > bitset->_rep.size()) {
		i = _rep.size();

		do {
			if (0ul != _rep[--i])
				return 1;
		} while (i > bitset->_rep.size());
	}
	else if (_rep.size() < bitset->_rep.size()) {
		i = bitset->_rep.size();

		do {
			if (0ul != bitset->_rep[--i])
				return -1;
		} while (_rep.size() < i);
	}
	else
		i = _rep.size();
		
	while (0 <  i--)
		if (_rep[i] < bitset->_rep[i])
			return -1;
		else if (_rep[i] > bitset->_rep[i])
			return 1;

	return 0;
}

EqcsBitsetRep *EqcsBitsetRep::not() const
{
	EqcsBitsetRep *ret = new EqcsBitsetRep(_nbits);

	for (int i = 0; i < _rep.size(); ++i) {
		ret->_rep[i] = ~_rep[i];

		if (i == (_rep.size() - 1) && 0 != (_nbits % BITS_PER_SLOT))
			ret->_rep[i] &= ~0ul >> (BITS_PER_SLOT - (_nbits % BITS_PER_SLOT));
	}

	return ret;
}

EqcsBitsetRep *EqcsBitsetRep::and(const EqcsBitsetRep *bitset) const
{
	EqcsBitsetRep *ret = new EqcsBitsetRep(max(_nbits, bitset->_nbits));

	for (int i = 0; i < min(_rep.size(), bitset->_rep.size()); ++i)
		ret->_rep[i] = _rep[i] & bitset->_rep[i];

	return ret;
}

EqcsBitsetRep *EqcsBitsetRep::or(const EqcsBitsetRep *bitset) const
{
	EqcsBitsetRep *ret = new EqcsBitsetRep(max(_nbits, bitset->_nbits));

	for (int i = 0; i < min(_rep.size(), bitset->_rep.size()); ++i)
		ret->_rep[i] = _rep[i] | bitset->_rep[i];

	return ret;
}


EqcsBitsetRep *EqcsBitsetRep::rsh(int distance) const
{
	int offset = distance / BITS_PER_SLOT;
	int delta = distance % BITS_PER_SLOT;
	slot_t carry = 0ul;
	EqcsBitsetRep *ret = new EqcsBitsetRep(_nbits);

	for (int i = _rep.size() - 1; i - offset >= 0; --i) {
		ret->_rep[i - offset] = (_rep[i] >> delta) | carry;
		carry = _rep[i] << (BITS_PER_SLOT - delta);
	}

	return ret;
}

EqcsBitsetRep *EqcsBitsetRep::lsh(int distance) const
{
	int offset = distance / BITS_PER_SLOT;
	int delta = distance % BITS_PER_SLOT;
	slot_t carry = 0ul;
	EqcsBitsetRep *ret = new EqcsBitsetRep(offset*BITS_PER_SLOT + _nbits);

	for (int i = 0; i < _rep.size(); ++i) {
		ret->_rep[i + offset] = (_rep[i] << delta) | carry;
		carry = _rep[i] >> (BITS_PER_SLOT - delta);
	}

	return ret;
}

ostream &EqcsBitsetRep::put_bit(ostream &os, int i) const
{
	if (i < _nbits) {
		int offs = i / BITS_PER_SLOT;
		slot_t mask = 1ul << (i % BITS_PER_SLOT);

		os.put(_rep[offs] & mask ? '1' : '0');
	}
	else
		os.put('0');
}

ostream &EqcsBitsetRep::put(ostream &os) const
{
	int i = _nbits;

	while (0 < i--)
		put_bit(os, i);

	return os;
}

ostream &EqcsBitsetRep::put_ket(ostream &os) const
{
	os.put('|');
	put(os);
	os.put('>');

	return os;
}

ostream &EqcsBitsetRep::put_ket(ostream &os, int n) const
{
	os.put('|');

	while (nbits() < n--)
		os.put('0');

	put(os);
	os.put('>');

	return os;
}

ostream &EqcsBitsetRep::put_bits(ostream &os, int nregs, int regwidth) const
{
	EqcsSep sep(">|");
	int n = nregs * regwidth;

	os.put('|');

	while (0 < n) {
		if (0 == (n % regwidth))
			os << sep;

		put_bit(os, --n);
	}

	os.put('>');

	return os;
}

ostream &EqcsBitsetRep::put_nums(ostream &os, int nregs, int regwidth) const
{
	EqcsSep sep(">|");
	int n = nregs * regwidth;
	int i = 0;

	os.put('|');

	while (0 < n) {
		if (0 == (n % regwidth))
			os << sep;

		if (test(--n))
			i |= 1 << (n % regwidth);

		if (0 == (n % regwidth)) {
			os << i;
			i = 0;
		}
	}

	os.put('>');

	return os;
}

int EqcsBitsetRep::dim(int nbits)
{
	int d = nbits / BITS_PER_SLOT;

	return 0 == nbits % BITS_PER_SLOT ? d : d + 1;
}
